#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/error.h"

/*
 * Known to devscsi.c.
 */
int scsidebugs[8] = {0};
int scsiownid = 7;

extern Conf	conf;

typedef struct Salloc	Salloc;

struct Salloc
{
	int	n;
	Lock	llock;
	QLock	rlock;
	Rendez	br;
	Scsibuf *avail;
};

static Salloc	salloc;

static int
isavail(void *p)
{
	return ((Salloc *)p)->avail != 0;
}

void
initscsi(void)
{
}

/*
 * Quick hack. Need to do a better job of dynamic initialisation
 * for machines with peculiar memory/cache restictions.
 * Also, what about 16Mb address limit on the Adaptec?
 */
static ulong bufdatasize;

void
scsibufreset(ulong datasize)
{
	Scsibuf *b;

	bufdatasize = datasize;
	if(conf.topofmem > 16*MB){
		for(salloc.n=0; salloc.n<4; salloc.n++){
			b = xalloc(sizeof(*b));
			b->virt = xalloc(bufdatasize);
			b->phys = (void *)(PADDR(b->virt));
			b->next = salloc.avail;
			salloc.avail = b;
			print("scsibuf: 0x%.6lux\n", b->phys);
			if((ulong)b->phys > 16*MB)
				panic("scsibuf");
		}
	}
}

Scsibuf *
scsibuf(void)
{
	Scsibuf *b;

	if(salloc.n <= 0){
		b = smalloc(sizeof(*b));
		b->virt = smalloc(bufdatasize);
		b->phys = (void *)(PADDR(b->virt));
	}else for(;;){
		lock(&salloc.llock);
		if(b = salloc.avail){	/* assign = */
			salloc.avail = b->next;
			unlock(&salloc.llock);
			break;
		}else{
			unlock(&salloc.llock);
			qlock(&salloc.rlock);
			if(waserror()){
				qunlock(&salloc.rlock);
				continue;
			}
			sleep(&salloc.br, isavail, &salloc);
			poperror();
			qunlock(&salloc.rlock);
		}
	}
	return b;
}

void
scsifree(Scsibuf *b)
{
	if(salloc.n <= 0){
		free(b->virt);
		free(b);
	}else{
		lock(&salloc.llock);
		b->next = salloc.avail;
		salloc.avail = b;
		unlock(&salloc.llock);
		if(salloc.br.p)
			wakeup(&salloc.br);
	}
}

extern int (*aha1542reset(void))(Scsi*, int);
extern int (*ultra14freset(void))(Scsi*, int);

static int (*exec)(Scsi*, int);

int
scsiexec(Scsi *p, int rflag)
{
	if(exec == 0)
		error(Enonexist);
	return (*exec)(p, rflag);
}

void
resetscsi(void)
{
	if(exec = aha1542reset())
		return;
	exec = ultra14freset();
}
