#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"../port/error.h"
#include	"devtab.h"
#include	"io.h"

/*
 * Ricoh scanner.  Gutted version of devwren.
 */

#define DATASIZE	(2*1024*1024)	/* default ricohsize */

ulong	ricohsize;
ulong	ricohn;
Scsibuf	ricohbuf;

enum {
	/* file types */
	Qdir=		0,
	Qdata=		1,
};

static int
ricohgen(Chan *c, Dirtab *tab, long ntab, long s, Dir *dirp)
{
	char name[NAMELEN+4];

	USED(tab, ntab);
	if(s != 0)
		return -1;
	sprint(name, "ricoh%d", c->dev);
	devdir(c, (Qid){Qdata, 0}, name, ricohsize, eve, 0444, dirp);
	return 1;
}

void
ricohreset(void)
{
	char *p;

	if(p = getconf("ricohsize"))	/* assign = */
		ricohsize = strtoul(p, 0, 0);
	else
		ricohsize = DATASIZE;
	ricohbuf.virt = xalloc(ricohsize);
	ricohbuf.phys = (void *)(PADDR(ricohbuf.virt));
	ricohbuf.next = 0;
}

void
ricohinit(void)
{
}

/*
 *  param is #ρ<target>
 */
Chan*
ricohattach(char *param)
{
	Chan *c;

	if(param == 0 || param[0] < '0' || param[0] > '7' || param[1])
		error(Ebadarg);
	c = devattach(L'ρ', param);
	c->dev = param[0] - '0';
	return c;
}

Chan*
ricohclone(Chan *c, Chan *nc)
{
	return devclone(c, nc);
}

int
ricohwalk(Chan *c, char *name)
{
	return devwalk(c, name, 0, 0, ricohgen);
}

void
ricohstat(Chan *c, char *dp)
{
	devstat(c, dp, 0, 0, ricohgen);
}

Chan*
ricohopen(Chan *c, int omode)
{
	return devopen(c, omode, 0, 0, ricohgen);
}

void
ricohcreate(Chan *c, char *name, int omode, ulong perm)
{
	USED(c, name, omode, perm);
	error(Eperm);
}

void
ricohclose(Chan *c)
{
	USED(c);
}

void
ricohremove(Chan *c)
{
	USED(c);
	error(Eperm);
}

void
ricohwstat(Chan *c, char *dp)
{
	USED(c, dp);
	error(Eperm);
}

long
ricohread(Chan *c, char *a, long n, ulong offset)
{
	int status;
	Scsi *cmd;

	if(c->qid.path == CHDIR)
		return devdirread(c, a, n, 0, 0, ricohgen);
	if(offset == 0){
		cmd = scsicmd(c->dev<<3, ScsiExtread, &ricohbuf, ricohsize);
		if(waserror()){
			qunlock(cmd);
			nexterror();
		}
		cmd->cmdblk[6] = ricohsize>>16;
		cmd->cmdblk[7] = ricohsize>>8;
		cmd->cmdblk[8] = ricohsize;
		status = scsiexec(cmd, ScsiIn);
		ricohn = cmd->data.ptr - cmd->data.base;
		poperror();
		qunlock(cmd);
		kprint("ricoh: n=%d, status=0x%.4ux\n", ricohn, status);
		if(ricohn <= 0 && status == 0x6002){
			ricohn = 0;
			error(Eio);
		}
	}
	if(offset >= ricohn)
		return 0;
	if(offset+n > ricohn)
		n = ricohn-offset;
	memmove(a, ((char*)ricohbuf.virt)+offset, n);
	return n;
}

long
ricohwrite(Chan *c, char *a, long n, ulong offset)
{
	USED(c, a, n, offset);
	error(Eperm);
	return 0;
}
